/**
 * Quaternion class.
 *
 * Quaternions are a mathematical entity, similar to complex numbers.
 * Each quaternion has 3 "imaginary" parts and one "real" part.
 * There are rules to add, subtract, multiply and divide quaternions
 * in a consistent manner.
 *
 * Written by Robert Osada, March 1999.
 **/
#ifndef __QUATERNION_H__
#define __QUATERNION_H__

#include "Misc.h"

struct Quat
{
  static const Quat ZERO, ONE;

  // create object
  Quat (float x_=0, float y_=0, float z_=0, float w_=0) : x(x_), y(y_), z(z_), w(w_) {}
  Quat (const Vector3f &v)                              : x(v.x), y(v.y), z(v.z), w(0) {}

  // norm
  float norm () const { return x*x + y*y + z*z + w*w; }
  Quat  unit () const { return (*this) / Sqrt(norm()); }

  // conjugate
  Quat conj () const { return Quat(-x, -y, -z, w); }

  // operators
  bool  operator== (const Quat &r) const { return x==r.x && y==r.y && z==r.z && w==r.w; }
  Quat  operator+  (const Quat &r) const { return Quat(x+r.x, y+r.y, z+r.z, w+r.w); }
  Quat& operator+= (const Quat &r)       { x+=r.x; y+=r.y; z+=r.z; w+=r.w; return *this; }
  Quat  operator-  () const              { return Quat(-x, -y, -z, -w); }
  Quat  operator-  (const Quat &r) const { return Quat(x-r.x, y-r.y, z-r.z, w-r.w); }
  Quat& operator-= (const Quat &r)       { x-=r.x; y-=r.y; z-=r.z; w-=r.w; return *this; }
  Quat  operator*  (float s) const       { return Quat(x*s, y*s, z*s, w*s); }
  Quat& operator*= (float s)             { x*=s; y*=s; z*=s; w*=s; return *this; }
  Quat  operator/  (float s) const       { assert(s); return (*this)* (1/s); }
  Quat& operator/= (float s)             { assert(s); return (*this)*=(1/s); }
  Quat  operator*  (const Quat &r) const;
  Quat& operator*= (const Quat &r);
  Quat  operator/  (const Quat &r) const { return (*this) *  r.conj() /  r.norm(); }
  Quat& operator/= (const Quat &r)       { return (*this) *= r.conj() /= r.norm(); }

  // components
  float x, y, z, w;
};

inline float Dot (const Quat &l, const Quat &r)
{
  return l.x*r.x + l.y*r.y + l.z*r.z + l.w*r.w;
}

#endif
